#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <zlib.h>

#include "pktlib.h"
#include "pktbuf.h"
#include "pcap.h"

#define SPECIAL_COMPRESS_PROTO_FLAG 0x40

static pktbuf_t decompress_ip(pktbuf_t pktbuf)
{
  char *p;
  struct ip *iphdr;
  int hdrsize, paysize;
  int offset = 0, woffset = 0, size, s, ret;
  z_stream z;
  char ibuf[32*1024];
  char obuf[32*1024];
  char wbuf[64*1024];

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return pktbuf;

  iphdr = (struct ip *)p;
  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (!(iphdr->ip_p & SPECIAL_COMPRESS_PROTO_FLAG))
    return pktbuf;

  if (pktbuf_get_size(pktbuf) < hdrsize + paysize)
    return pktbuf;

  size = paysize;

  z.zalloc = Z_NULL;
  z.zfree = Z_NULL;
  z.opaque = Z_NULL;

  z.next_in = Z_NULL;
  z.avail_in = 0;
  ret = inflateInit(&z);
  if (ret != Z_OK)
    return pktbuf;

  z.next_out = (Bytef *)obuf;
  z.avail_out = sizeof(obuf);

  do {
    if (z.avail_in == 0) {
      s = sizeof(ibuf);
      if (s > size) s = size;
      memcpy(ibuf, p + hdrsize + offset, s);
      offset += s;
      size -= s;
      z.next_in = (Bytef *)ibuf;
      z.avail_in = s;
    }

    ret = inflate(&z, Z_NO_FLUSH);
    if ((ret != Z_OK) && (ret != Z_STREAM_END))
      return pktbuf;

    if ((z.avail_out == 0) || (ret == Z_STREAM_END)) {
      s = sizeof(obuf) - z.avail_out;
      memcpy(wbuf + woffset, obuf, s);
      woffset += s;
      z.next_out = (Bytef *)obuf;
      z.avail_out = sizeof(obuf);
    }
  } while (ret != Z_STREAM_END);

  if (inflateEnd(&z) != Z_OK)
    return pktbuf;

  paysize = woffset;

  memcpy(p + hdrsize, wbuf, paysize);

  pktbuf_set_size(pktbuf, hdrsize + paysize);

  iphdr->ip_p &= ~SPECIAL_COMPRESS_PROTO_FLAG;
  iphdr->ip_len = htons(hdrsize + paysize);

  return pktbuf;
}

int main(int argc, char *argv[])
{
  pktbuf_t pktbuf;
  int size, linktype, origsize;
  struct timeval tm;
  struct ether_header *ehdr;

  argc--; argv++;

  pkt_pcap_read(stdin, NULL, 0, &linktype, NULL, NULL); /* 󥯥פμ */

  while (1) {
    pktbuf = pktbuf_create(pktlib_linktype_offset(linktype));
    size = pkt_pcap_read(stdin, pktbuf_get_header(pktbuf), 80 * 1024,
			 &linktype, &origsize, &tm); /* pcap */
    if (size < 0)
      break;
    pktbuf_set_size(pktbuf, size);

    ehdr = (struct ether_header *)pktbuf_get_header(pktbuf);

    if (ntohs(ehdr->ether_type) == ETHERTYPE_IP) {
      pktbuf_delete_header(pktbuf, ETHER_HDR_LEN);
      pktbuf = decompress_ip(pktbuf);
      pktbuf = pktbuf_checksum_correct_ip(pktbuf);
      pktbuf_add_header(pktbuf, ETHER_HDR_LEN);
    }

    pkt_pcap_write(stdout, pktbuf_get_header(pktbuf), pktbuf_get_size(pktbuf),
                   linktype, pktbuf_get_size(pktbuf), &tm); /* pcap */
    fflush(stdout);

    pktbuf_destroy(pktbuf); /* ѥåȤѴ */
  }

  return 0;
}
