#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

#include "pktlib.h"
#include "pktbuf.h"
#include "arplib.h"

static char *make_srcmacaddr(in_addr_t ipaddr)
{
  static char macaddr[ETHER_ADDR_LEN];
  ipaddr = htonl(ipaddr);
  macaddr[0] = 0x00;
  macaddr[1] = 0x11;
  memcpy(&macaddr[2], &ipaddr, sizeof(in_addr_t));
  return macaddr;
}

static in_addr_t proc_ip(pktbuf_t pktbuf, in_addr_t my_ipaddr)
{
  char *p;
  struct ip *iphdr;
  struct icmp *icmphdr;
  int hdrsize, paysize;

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return 0;

  iphdr = (struct ip *)p;
  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (iphdr->ip_p != IPPROTO_ICMP)
    return 0;
  if (ntohl(iphdr->ip_dst.s_addr) != my_ipaddr)
    return 0;

  if (pktbuf_get_size(pktbuf) < hdrsize + paysize)
    return 0;

  if (paysize < sizeof(*icmphdr))
    return 0;

  icmphdr = (struct icmp *)(p + hdrsize);
  if (icmphdr->icmp_type != ICMP_ECHO)
    return 0;

  fprintf(stderr, "RECV ICMP Echo       %-15s", inet_ntoa(iphdr->ip_src));
  fprintf(stderr, " -> %-15s\n", inet_ntoa(iphdr->ip_dst));

  icmphdr->icmp_type = ICMP_ECHOREPLY;
  icmphdr->icmp_cksum = 0;

  iphdr->ip_dst.s_addr = iphdr->ip_src.s_addr;
  iphdr->ip_src.s_addr = htonl(my_ipaddr);
  iphdr->ip_sum = 0;

  fprintf(stderr, "SEND ICMP Echo Reply %-15s", inet_ntoa(iphdr->ip_src));
  fprintf(stderr, " -> %-15s\n", inet_ntoa(iphdr->ip_dst));

  return ntohl(iphdr->ip_dst.s_addr);
}

static in_addr_t proc_ttl(pktbuf_t pktbuf)
{
  char *p;
  struct ip *iphdr;

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return 0;

  iphdr = (struct ip *)p;
  if (iphdr->ip_ttl <= 1)
    return 0;

  iphdr->ip_ttl--;

  return ntohl(iphdr->ip_dst.s_addr);
}

struct ip_iface {
  pktif_t pktif;
  char macaddr[ETHER_ADDR_LEN];
  in_addr_t ipaddr;
  in_addr_t ipmask;
  arptbl_t arptbl;
};

struct ip_route {
  in_addr_t ipaddr;
  in_addr_t ipmask;
  in_addr_t gateway;
  struct ip_iface *ifp;
};

int main(int argc, char *argv[])
{
  pktif_t pktif;
  pktbuf_t pktbuf;
  int ifnum = 0, rtnum = 0;
  in_addr_t dstaddr;
  struct in_addr in;
  struct ether_header *ehdr;
  struct ip *iphdr;
  struct ip_iface iflist[16], *ifp = NULL;
  struct ip_route rtlist[16], *rtp = rtlist, *found;

  argc--; argv++;
  /* 󥿡ե */
  for (ifp = iflist; argc > 2; argc -= 3, argv += 3) {
    if (!strcmp(argv[0], "--"))
      break;
    /* 󥿡ե򥪡ץ󤹤(ȯѥåȤϼʤ) */
    ifp->pktif = pktif_open(argv[0], PKTIF_OPEN_FLAG_RECV_NOTSENT, 0);
    ifp->ipaddr = ntohl(inet_addr(argv[1])); /* IPɥ쥹 */
    ifp->ipmask = ntohl(inet_addr(argv[2])); /* ͥåȥޥ */
    memcpy(ifp->macaddr, make_srcmacaddr(ifp->ipaddr), ETHER_ADDR_LEN);
    ifp->arptbl = arplib_table_create(ifp->ipaddr, ifp->macaddr);

    /* ͥƥåɡ롼Ȥ */
    rtp->ipaddr = ifp->ipaddr & ifp->ipmask;
    rtp->ipmask = ifp->ipmask;
    rtp->gateway = 0;
    rtp->ifp = ifp;
    rtnum++;
    rtp++;

    ifnum++;
    ifp++;
  }

  argc--; argv++;
  /* ϩ */
  for (; argc > 2; argc -= 3, argv += 3) {
    rtp->ipaddr  = ntohl(inet_addr(argv[0])); /* IPɥ쥹 */
    rtp->ipmask  = ntohl(inet_addr(argv[1])); /* ͥåȥޥ */
    rtp->gateway = ntohl(inet_addr(argv[2])); /* ž */
    rtp->ifp = NULL;
    rtnum++;
    rtp++;
  }

  fprintf(stderr, "Interfaces:\n");
  for (ifp = iflist; ifp < &iflist[ifnum]; ifp++) {
    fprintf(stderr, "\t%s", pktif_get_name(ifp->pktif));
    in.s_addr = htonl(ifp->ipaddr);
    fprintf(stderr, " %s", inet_ntoa(in));
    in.s_addr = htonl(ifp->ipmask);
    fprintf(stderr, "/%s", inet_ntoa(in));
    fprintf(stderr, "\n");
  }

  fprintf(stderr, "Routes:\n");
  for (rtp = rtlist; rtp < &rtlist[rtnum]; rtp++) {
    in.s_addr = htonl(rtp->ipaddr);
    fprintf(stderr, "\t%s", inet_ntoa(in));
    in.s_addr = htonl(rtp->ipmask);
    fprintf(stderr, "/%s", inet_ntoa(in));
    if (rtp->ifp) {
      fprintf(stderr, " -> %s\n", pktif_get_name(rtp->ifp->pktif));
    } else {
      in.s_addr = htonl(rtp->gateway);
      fprintf(stderr, " -> %s\n", inet_ntoa(in));
    }
  }

  while (1) {
    pktif = pktlib_iflist_select(-1); /* ʣ󥿡եԤ */
    if (pktif == NULL) continue;

    pktbuf = pktbuf_recv(pktif, 0); /* ѥåȤ */
    if (pktbuf == NULL) continue;

    /* 󥿡ե򸡺 */
    for (ifp = iflist; ifp < &iflist[ifnum]; ifp++) {
      if (pktif == ifp->pktif)
	break;
    }
    if (ifp == &iflist[ifnum])
      goto discard;

    ehdr = (struct ether_header *)pktbuf_get_header(pktbuf);

    /* MACɥ쥹åβݤȽǤ */
    if (!(ehdr->ether_dhost[0] & 1) &&
	memcmp(ehdr->ether_dhost, ifp->macaddr, ETHER_ADDR_LEN))
      goto discard;

    /* ARPν */
    if (ntohs(ehdr->ether_type) == ETHERTYPE_ARP) {
      pktbuf = arplib_packet_proc(ifp->arptbl, pktbuf);
      pktbuf = pktbuf_send_queue(ifp->pktif, pktbuf); /* ʣѥåȤ */
      goto discard;
    }

    if (ntohs(ehdr->ether_type) != ETHERTYPE_IP)
      goto discard;

    pktbuf_delete_header(pktbuf, ETHER_HDR_LEN);

    iphdr = (struct ip *)pktbuf_get_header(pktbuf);

    fprintf(stderr, "RECV IP Packet %s %-15s",
	    pktif_get_name(ifp->pktif), inet_ntoa(iphdr->ip_src));
    fprintf(stderr, " -> %-15s\n", inet_ntoa(iphdr->ip_dst));

    if ((ntohl(iphdr->ip_dst.s_addr) & 0xF0000000) == 0xE0000000)
      goto discard; /* IPޥ㥹Ȥ˴ */

    /* ȰΥѥåȤν */
    dstaddr = 0;
    for (ifp = iflist; ifp < &iflist[ifnum]; ifp++) {
      if ((ntohl(iphdr->ip_dst.s_addr) | ifp->ipmask) == 0xFFFFFFFF)
	goto discard; /* IP֥ɥ㥹Ȥ˴ */
      dstaddr = proc_ip(pktbuf, ifp->ipaddr); /* ICMP Echo ʤб */
      if (dstaddr)
	break;
    }
    if (dstaddr == 0) {
      dstaddr = proc_ttl(pktbuf);
      if (dstaddr == 0)
	goto discard;
    }
    iphdr = (struct ip *)pktbuf_get_header(pktbuf);
    dstaddr = ntohl(iphdr->ip_dst.s_addr);

    /* ϩ򸡺 */
    while (1) {
      found = NULL;
      for (rtp = rtlist; rtp < &rtlist[rtnum]; rtp++) {
	if ((dstaddr & rtp->ipmask) == (rtp->ipaddr & rtp->ipmask)) {
	  /* ʣηϩҥåȤˤϡ󥲥ȥޥåΥ롼Ƿ */
	  if ((found == NULL) || (rtp->ipmask > found->ipmask))
	    found = rtp;
	}
      }
      if (found == NULL) /* ϩĤʤ */
	break;
      ifp = found->ifp;
      if (ifp) /* Υ󥿡եꤷ */
	goto sendpkt;
      dstaddr = found->gateway; /* žФƷϩƸ */
    }

    goto discard;

  sendpkt:
    in.s_addr = htonl(dstaddr);
    fprintf(stderr, "SEND IP Packet %s %-15s\n",
	    pktif_get_name(ifp->pktif), inet_ntoa(in));
    pktbuf = pktbuf_checksum_correct_ip(pktbuf);
    pktbuf = arplib_packet_make(ifp->arptbl, pktbuf, ETHERTYPE_IP, dstaddr);
    pktbuf = pktbuf_send_queue(ifp->pktif, pktbuf); /* ʣѥåȤ */

  discard:
    pktbuf_destroy_queue(pktbuf);
  }

  return 0;
}
